import enum


class UserType(enum.Enum):
    AndroidUser = 2


class UserChannel(enum.Enum):
    Mihoyo = "mihoyo"


class UserCGGameBiz(enum.Enum):
    CN = "hk4e_cn"


class UserOpBiz(enum.Enum):
    CN = "clgm_cn"


class UserCps(enum.Enum):
    Mihoyo = "mihoyo"


class User:
    def __init__(
            self,
            combo_token: str,
            sys_version: str,
            device_id: str,
            device_name: str,
            device_model: str,
            *,
            user_type: UserType = UserType.AndroidUser,
            channel: UserChannel = UserChannel.Mihoyo,
            cg_game_biz: UserCGGameBiz = UserCGGameBiz.CN,
            op_biz: UserOpBiz = UserOpBiz.CN,
            cps: UserCps = UserCps.Mihoyo,
            language: str = "zh-cn"
    ):
        self._combo_token = combo_token
        self._sys_version = sys_version
        self._device_id = device_id
        self._device_name = device_name
        self._device_model = device_model
        self._user_type = user_type
        self._channel = channel
        self._cg_game_biz = cg_game_biz
        self._op_biz = op_biz
        self._cps = cps
        self._language = language

    def __str__(self) -> str:
        return (f"User[combo_token: {self._combo_token}, sys_version: {self._sys_version}, "
                f"device_id: {self._device_id}, device_name: {self._device_name}, "
                f"device_model: {self._device_model}, user_type: {self._user_type.name}, "
                f"channel: {self._channel.name}, cg_game_biz: {self._cg_game_biz.name}, "
                f"op_biz: {self._op_biz.name}, cps: {self._cps}, language: {self._language}]")

    def __repr__(self) -> str:
        return (f"User(combo_token={self._combo_token}, sys_version={self._sys_version}, device_id={self._device_id}, "
                f"device_name={self._device_name}, device_model={self._device_model}, "
                f"user_type={self._user_type}, channel={self._channel}, "
                f"cg_game_biz={self._cg_game_biz}, op_biz={self._op_biz}, "
                f"cps={self._cps}, language={self._language})")

    @property
    def combo_token(self) -> str:
        return self._combo_token

    @property
    def sys_version(self) -> str:
        return self._sys_version

    @property
    def device_id(self) -> str:
        return self._device_id

    @property
    def device_name(self) -> str:
        return self._device_name

    @property
    def device_model(self) -> str:
        return self.device_model

    @property
    def user_type(self) -> UserType:
        return self._user_type

    @property
    def channel(self) -> UserChannel:
        return self._channel

    @property
    def cps(self) -> UserCps:
        return self._cps

    @property
    def cg_game_biz(self) -> UserCGGameBiz:
        return self.cg_game_biz

    @property
    def language(self) -> str:
        return self._language

    @property
    def op_biz(self) -> UserOpBiz:
        return self._op_biz

    @property
    def header(self) -> dict:
        """
        用户层的请求头
        :return: 发出 Web 请求时的 `Headers`
        """
        return {
            "x-rpc-combo_token": self._combo_token,
            "x-rpc-client_type": str(self._user_type.value),
            "x-rpc-sys_version": self._sys_version,
            "x-rpc-channel": self._channel.value,
            "x-rpc-cps": self._cps.value,
            "x-rpc-cg_game_biz": self._cg_game_biz.value,
            "x-rpc-device_id": self._device_id,
            "x-rpc-device_name": self._device_name,
            "x-rpc-device_model": self._device_model,
            "x-rpc-language": self._language,
            "x-rpc-op_biz": self._op_biz.value
        }



import enum
import typing
import httpx

from typing import Optional, List

from ._api import APIStatic, APICloudGame
from ._wallet import WalletData
from ._user import User
from ._exception import WebRequestError, APIError
from ._notification import Notification, NotificationStatus, NotificationType

T = typing.TypeVar("T", bound="Client")


class ClientStatus(enum.Enum):
    # UNOPENED:
    #   The client has been instantiated, but has not been used to send a web request,
    #   or been opened by entering the context of a `with` block.
    UNOPENED = 0
    # OPENED:
    #   The client has either sent a web request, or is within a `with` block.
    OPENED = 1
    # CLOSED:
    #   The client has either exited the `with` block, or `close()` has been called explicitly.
    CLOSED = 2


class Client:
    def __init__(self: T):
        self._client = httpx.Client()
        self._status = ClientStatus.UNOPENED
        version_rep = self._client.get(APIStatic.VERSION)
        self._version = version_rep.json()["data"]["game"]["latest"]["version"]

    def __repr__(self) -> str:
        return f"Client()"

    def __str__(self) -> str:
        return f"Client[version: {self.version}, is_closed: {self.is_closed}, status: {self.status.name}]"

    def __enter__(self: T) -> T:
        if self._status != ClientStatus.UNOPENED:
            msg = {
                ClientStatus.OPENED: "Cannot open a client instance more than once.",
                ClientStatus.CLOSED: "Cannot reopen a client instance, once it has been closed.",
            }[self._status]
            raise RuntimeError(msg)

        self._status = ClientStatus.OPENED
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._status = ClientStatus.CLOSED
        self._client.close()

    def close(self) -> None:
        self._status = ClientStatus.CLOSED
        self._client.close()

    @property
    def is_closed(self) -> bool:
        return self._status == ClientStatus.CLOSED

    @property
    def status(self) -> ClientStatus:
        return self._status

    @property
    def version(self) -> str:
        return self._version

    def _get_common_headers(self) -> dict:
        return {
            "x-rpc-app_version": self._version,
            "x-rpc-app_id": "1953439974",
            "x-rpc-vendor_id": "1",
            "Referer": "https://app.mihoyo.com"
        }

    def _web_get(self, user: User, url: str, *, params: Optional[dict] = None) -> httpx.Response:
        if self._status == ClientStatus.CLOSED:
            raise RuntimeError("Cannot send a request, as the client has been closed.")

        self._status = ClientStatus.OPENED

        headers = self._get_common_headers()
        headers.update(user.header)
        resp = self._client.get(url, headers=headers, params=params)
        if resp.status_code != 200:
            raise WebRequestError(f"Status code: {resp.status_code}")
        return resp

    def get_wallet(self, user: User) -> WalletData:
        resp = self._web_get(user, APICloudGame.WALLET)
        resp_data = resp.json()
        if resp_data["retcode"] != 0:
            raise APIError(f"Retcode: {resp_data['retcode']}, Message: {resp_data['message']}")
        return WalletData(resp_data["data"])

    def get_notifications(
            self,
            user: User,
            *,
            status: Optional[NotificationStatus] = None,
            ntype: Optional[NotificationType] = None
    ) -> List[Notification]:
        params = {}
        if status is not None:
            params["status"] = status.value
        if ntype is not None:
            params["type"] = ntype.value

        resp = self._web_get(user, APICloudGame.NOTIFICATION, params=params)
        resp_data = resp.json()
        if resp_data["retcode"] != 0:
            raise APIError(f"Retcode: {resp_data['retcode']}, Message: {resp_data['message']}")
        notifications = []
        for data in resp_data["data"]["list"]:
            notifications.append(Notification(data))
        return notifications



class WebRequestError(RuntimeError):
    def __init__(self, message):
        self.message = message


class APIError(RuntimeError):
    def __init__(self, message):
        self.message = message


class UndefinedNameWarning(RuntimeWarning):
    def __init__(self, message):
        self.message = message



# Entities
from ._client import Client
from ._user import User

# Enums
from ._user import UserType, UserChannel, UserCGGameBiz, UserOpBiz, UserCps
from ._notification import NotificationType, NotificationStatus



API_CLOUDGAME = "https://api-cloudgame.mihoyo.com"
SDK_STATIC = "https://sdk-static.mihoyo.com"


class APICloudGame:
    WALLET = API_CLOUDGAME + "/hk4e_cg_cn/wallet/wallet/get"
    NOTIFICATION = API_CLOUDGAME + "/hk4e_cg_cn/gamer/api/listNotifications"


class APIStatic:
    VERSION = SDK_STATIC + "/hk4e_cn/mdk/launcher/api/resource?key=eYd89JmJ&launcher_id=18"



import datetime


class WalletData:
    """
    用户的钱包数据
    """

    class _Coin:
        """
        米云币
        """

        def __init__(self, data: dict):
            self._data = data

        def __repr__(self) -> str:
            return f"WalletData._Coin({self._data})"

        def __str__(self) -> str:
            return f"WalletData.Coin[coin: {self.coin}, free: {self.free}, limit: {self.limit}]"

        @property
        def coin(self) -> int:
            """
            米云币数量
            """
            return int(self._data["coin_num"])

        @property
        def free(self) -> int:
            """
            免费米云币数量
            """
            return int(self._data["free_coin_num"])

        @property
        def limit(self) -> int:
            """
            米云币上限
            """
            return int(self._data["coin_limit"])

    class _FreeTime:
        """
        免费时长
        """

        def __init__(self, data: dict):
            self._data = data

        def __repr__(self) -> str:
            return f"WalletData._Freetime({self._data})"

        def __str__(self) -> str:
            return (f"WalletData.Freetime["
                    f"sent: {self.sent}, total: {self.total}, "
                    f"limit: {self.limit}, overflow: {self.overflow}]")

        @property
        def sent(self) -> datetime.time:
            """
            每日登陆赠送的时长
            :return: 如果是首次请求(即签到动作)，则返回赠送的时长，否则为 `0`
            """
            send_free_time = int(self._data["send_freetime"])
            return datetime.time(send_free_time // 60, send_free_time % 60)

        @property
        def total(self) -> datetime.time:
            """
            总共的免费时长
            """
            free_time = int(self._data["free_time"])
            return datetime.time(free_time // 60, free_time % 60)

        @property
        def limit(self) -> datetime.time:
            """
            免费时长上限
            """
            limit = int(self._data["free_time_limit"])
            return datetime.time(limit // 60, limit % 60)

        @property
        def overflow(self) -> datetime.time:
            """
            溢出的时长
            :return: 如果是首次请求(即签到动作), 则返回本次签到溢出的时长
            """
            over = int(self._data["over_freetime"])
            return datetime.time(over // 60, over % 60)

    class _PlayCard:
        """
        畅玩卡
        """

        def __init__(self, data: dict):
            self._data = data

        def __repr__(self) -> str:
            return f"WalletData._PlayCard(f{self._data})"

        def __str__(self) -> str:
            return f"WalletData.PlayCard[expire: {self.expire}, message: {self.message}, status: {self.status}]"

        @property
        def expire(self) -> str:
            """
            过期时间
            :return: 字符串形式的过期时间
            """
            return self._data["expire"]

        @property
        def message(self) -> str:
            """
            点击开通畅玩卡上面的信息
            :return: 提示信息
            """
            return self._data["msg"]

        @property
        def status(self) -> str:
            """
            状态信息
            """
            return self._data["short_msg"]

    def __init__(self, data: dict):
        self._data = data
        self._coin = self._Coin(data["coin"])
        self._free_time = self._FreeTime(data["free_time"])
        self._play_card = self._PlayCard(data["play_card"])

    def __repr__(self) -> str:
        return f"WalletData({self._data})"

    def __str__(self) -> str:
        return f"WalletData[Coin: {self.coin}, FreeTime: {self.free_time}, PlayCard: {self.play_card}]"

    @property
    def coin(self):
        """
        米云币
        """
        return self._coin

    @property
    def free_time(self):
        """
        免费时长
        """
        return self._free_time

    @property
    def play_card(self):
        """
        畅玩卡
        """
        return self._play_card

    @property
    def is_signin(self) -> bool:
        """
        是否是每日签到动作
        :return: 是则为真
        """
        # 这里的逻辑是如果给的免费时间为 0 就不是第一次上线了
        return self.free_time.sent != datetime.time(0, 0)



import warnings
import datetime
import enum

from ._exception import UndefinedNameWarning


class NotificationStatus(enum.Enum):
    UNREAD = "NotificationStatusUnread"
    READ = "NotificationStatusRead"
    UNDEFINED = "NotificationStatusUndefined"


class NotificationType(enum.Enum):
    POPUP = "NotificationTypePopup"
    UNDEFINED = "NotificationTypeUndefined"


class Notification:
    def __init__(self, data: dict):
        self._data = data

        if data["status"] not in [status.value for status in NotificationStatus]:
            self._status = NotificationStatus.UNDEFINED
            warnings.warn(f'The name {data["status"]} is undefined.', UndefinedNameWarning)
        else:
            self._status = NotificationStatus(data["status"])

        if data["type"] not in [ntype.value for ntype in NotificationType]:
            self._type = NotificationType.UNDEFINED
            warnings.warn(f'The name {data["type"]} is undefined.', UndefinedNameWarning)
        else:
            self._type = NotificationType(data["type"])

    def __repr__(self) -> str:
        return f"Notification({self._data})"

    def __str__(self) -> str:
        return f"Notification<{self.msg}>"

    @property
    def id(self) -> str:
        return self._data["id"]

    @property
    def status(self) -> NotificationStatus:
        return self._status

    @property
    def type(self) -> NotificationType:
        return self._type

    @property
    def priority(self) -> int:
        return self._data["priority"]

    @property
    def source(self) -> str:
        return self._data["source"]

    @property
    def desc(self) -> str:
        return self._data["desc"]

    @property
    def msg(self) -> str:
        return self._data["msg"]

    @property
    def created_at(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(int(self._data["created_at"]))
